# Import core components
from pyw2v2.components.model_ctc import ModelCTC
from pyw2v2.components.metrics import Metrics
from pyw2v2.components.dataset_proc import DatasetPreprocessor

# Import external components
from pyw2v2.components.external.data_collator_ctc import DataCollatorCTCWithPadding

# Import submodules
import pyw2v2.utils as utils



from datasets import list_metrics, load_metric
from numpy.lib.arraysetops import isin


class Metrics:
    
    def __init__(self, *args):
        _all_metrics = list_metrics()
        self._metrics = {m: load_metric(m) for m in args if m in _all_metrics and isinstance(m, str)}
        
    def compute(self, predictions, references) -> dict:
        return {n: m.compute(predictions=predictions, references=references) for n, m in self._metrics.items()}
    
    def list(self) -> list:
        return list(self._metrics.keys())


import torch
import numpy as np
from easydict import EasyDict
from datasets import Dataset
from transformers import Wav2Vec2CTCTokenizer, Wav2Vec2FeatureExtractor, Wav2Vec2Processor, Wav2Vec2ForCTC
from transformers import TrainingArguments, Trainer

from pyw2v2.components.metrics import Metrics
from pyw2v2.components.external.data_collator_ctc import DataCollatorCTCWithPadding
from pyw2v2.components.dataset_proc import DatasetPreprocessor


class ModelCTC:

    def __init__(self, config: EasyDict):
        self._model = None
        self._metrics = None
        self._processor = None
        self._data_collator = None
        self._training_args = None
        self._dataset_preprocessor = None
        
        self._init_processor(config)
        self._init_metrics(config)
        self._init_data_collator()
        self._init_model(config)
        self._init_training_args(config)
        
    @property
    def processor(self):
        return self._processor
    
    @processor.setter
    def processor(self, var: Wav2Vec2Processor):
        self._processor = var

    @property
    def dataset_preprocessor(self):
        return self._dataset_preprocessor
    
    @processor.setter
    def dataset_preprocessor(self, var: DatasetPreprocessor):
        self._dataset_preprocessor = var

    def _init_processor(self, config: EasyDict):
        config.processor.tokenizer.vocab_file = config.common.vocab_file
        tokenizer = Wav2Vec2CTCTokenizer(**config.processor.tokenizer)
        feature_extractor = Wav2Vec2FeatureExtractor(**config.processor.feature_extractor)

        processor = Wav2Vec2Processor(feature_extractor=feature_extractor, tokenizer=tokenizer)
        processor.save_pretrained(config.common.model_path)
        self._processor = processor

    def _init_metrics(self, config: EasyDict):
        self._metrics = Metrics(*config.common.metrics)

    def _init_data_collator(self):
        self._data_collator = DataCollatorCTCWithPadding(processor=self._processor, padding=True)

    def _init_model(self, config: EasyDict):
        if not config.common.checkpoint_model:
            print(f"Loading pretrained model {config.common.pretrained_model}")
            config.model.pretrained_model_name_or_path = config.common.pretrained_model
            config.model.pad_token_id = self._processor.tokenizer.pad_token_id
            config.model.vocab_size = len(self._processor.tokenizer)
            self._model = Wav2Vec2ForCTC.from_pretrained(**config.model)
        else:
            print(f"Loading from checkpoint {config.common.checkpoint_model}")
            self._model = Wav2Vec2ForCTC.from_pretrained(config.common.checkpoint_model).to("cuda")
    
    def _init_training_args(self, config: EasyDict):
        config.training_args.output_dir = config.common.model_path
        self._training_args = TrainingArguments(**config.training_args)
        
    def _compute_metrics(self, pred):
        pred_logits = pred.predictions
        pred_ids = np.argmax(pred_logits, axis=-1)

        pred.label_ids[pred.label_ids == -100] = self._processor.tokenizer.pad_token_id
        pred_str = self._processor.batch_decode(pred_ids)
        label_str = self._processor.batch_decode(pred.label_ids, group_tokens=False)

        res = self._metrics.compute(predictions=pred_str, references=label_str)
        return res
    
    def train(self, train_set: Dataset, eval_set: Dataset):
        trainer = Trainer(
            model=self._model,
            data_collator=self._data_collator,
            args=self._training_args,
            compute_metrics=self._compute_metrics,
            train_dataset=train_set,
            eval_dataset=eval_set,
            tokenizer=self._processor.feature_extractor)
        trainer.train()

    def decode_sample(self, batch):
        input_dict = self._processor(batch["input_values"], return_tensors="pt", padding=True)
        logits = self._model(input_dict.input_values.to("cuda")).logits
        pred_ids = torch.argmax(logits, dim=-1)
        batch['decoded'] = self._processor.decode(pred_ids)
        return batch

    def decode(self, dataset: Dataset):
        return dataset.map(self.decode_sample)

    def decode_raw(self, path: str or list):
        if isinstance(path, str):
            sample = self._dataset_preprocessor.process_one(path, audio_only=True, make_labels=False)
            return self.decode_sample(sample)
        elif isinstance(path, list):
            samples = [self._dataset_preprocessor.process_one(p, audio_only=True, make_labels=False) for p in path]
            return [self.decode_sample(s) for s in samples]
        else:
            raise NotImplementedError



import librosa
import torchaudio
import numpy as np
from datasets import Dataset
from easydict import EasyDict
from transformers import Wav2Vec2Processor
# from datasets.utils.logging import set_verbosity_error
# set_verbosity_error()


class DatasetPreprocessor:
    
    def __init__(self, config: EasyDict, processor: Wav2Vec2Processor=None):
        self._text_column = config.text_column
        self._path_column = config.path_column
        self._remove_chars = config.remove_chars
        self._replace_chars = config.replace_chars
        self._sampling_rate = config.sampling_rate
        self._num_proc = config.n_jobs
        self._processor = processor
        
    @property
    def processor(self):
        return self._processor
    
    @processor.setter
    def processor(self, var: Wav2Vec2Processor):
        self._processor = var

    def _to_lower(self, batch, text_column=None):
        text_column = text_column if text_column is not None else self._path_column

        batch[self._text_column] = batch[text_column].lower() 
        return batch

    def _remove_characters(self, batch, text_column=None):
        text_column = text_column if text_column is not None else self._path_column

        for char in self._remove_chars:
            batch[text_column] = batch[text_column].replace(char, '')
        return batch
    
    def _replace_characters(self, batch, text_column=None):
        text_column = text_column if text_column is not None else self._path_column

        for k, v in self._replace_chars.items():
            batch[text_column] = batch[text_column].replace(k, v)
        return batch
    
    def _speech_file_to_array(self, batch, path_column=None):
        path_column = path_column if path_column is not None else self._path_column

        speech_array, sampling_rate = torchaudio.load(batch[path_column])
        batch["speech"] = speech_array[0].numpy()
        batch["sampling_rate"] = sampling_rate
        # if make_labels:
        #     batch["target_text"] = batch[self._text_column]
        return batch
    
    def _resample(self, batch, sampling_rate=None):
        sampling_rate = sampling_rate if sampling_rate is not None else self._sampling_rate

        batch["speech"] = librosa.resample(np.asarray(batch["speech"]), batch["sampling_rate"], sampling_rate)
        batch["sampling_rate"] = sampling_rate
        return batch

    def _prepare_batch(self, batch, make_labels=True):
        assert (
            len(set(batch["sampling_rate"])) == 1
        ), f"Make sure all inputs have the same sampling rate of {self._processor.feature_extractor.sampling_rate}."
        batch["input_values"] = self._processor(batch["speech"], sampling_rate=batch["sampling_rate"][0]).input_values
        if make_labels:
            with self._processor.as_target_processor():
                batch["labels"] = self._processor(batch[self._text_column]).input_ids
        return batch
    
    @staticmethod
    def dataset_head(dataset: Dataset, n: int) -> Dataset:
        return Dataset.from_dict(dataset[:n])

    def process(self, dataset: Dataset, n_samples: int=None, audio_only=False, make_labels=True) -> Dataset:
        if n_samples:
            dataset = self.dataset_head(dataset, n_samples)
        if not audio_only:
            dataset = dataset.map(self._remove_characters, num_proc=self._num_proc)
            dataset = dataset.map(self._replace_characters, num_proc=self._num_proc)
            dataset = dataset.map(self._to_lower, num_proc=self._num_proc)
        dataset = dataset.map(self._speech_file_to_array, num_proc=self._num_proc)
                            #   fn_kwargs={'make_labels': make_labels})
        dataset = dataset.map(self._resample, num_proc=self._num_proc)
        dataset = dataset.map(self._prepare_batch, batch_size=8, batched=True, num_proc=self._num_proc,
                              fn_kwargs={'make_labels': make_labels})
        return dataset
    
    def process_one(self, path: str, text: str=None, audio_only=False, make_labels=True, sampling_rate=None):
        if not text and not audio_only:
            raise NotImplementedError("text argument must be specified if audio_only=False!")
        
        data = {'path': path, 'text': text}
        
        if not audio_only:
            data = self._remove_characters(data, text_column='text')
            data = self._replace_characters(data, text_column='text')
            data = self._to_lower(data, text_column='text')
        data = self._speech_file_to_array(data, path_column='path')
        data = self._resample(data, sampling_rate=sampling_rate)
        data = self._prepare_batch(data, make_labels=make_labels)
        return data
            
        



import os
import yaml
import pandas as pd
from easydict import EasyDict
from pathlib import Path
from datasets import Dataset


def load_config(path):
    path = Path(path)
    if os.path.exists(path):
        with open(path) as f:
            config = EasyDict(yaml.safe_load(f))
    else:
        raise FileNotFoundError(f"No such file {path}")
    return config


def load_custom_dataset_commonvoice_format(path, split, path_column='path'):
    # TODO: add support for multiple split to be together. Example: train+validation
    dataset_path = Path(path) / (split + '.tsv')
    df = pd.read_csv(dataset_path, sep='\t')
    df[path_column] = [str((Path(path) / p).absolute()) for _, p in df[path_column].iteritems()]
    return Dataset.from_pandas(df)


from setuptools import setup, find_packages

VERSION = '0.0.1a3'
DESCRIPTION = 'Simple wav2vec2 wrapper'

with open("./README.md", 'r') as f:
    long_description = "\n" + "".join(f.readlines())

# Setting up
setup(
    name="pyw2v2",
    version=VERSION,
    author="vbrydik (Vitalii Brydinskyi)",
    author_email="<vbrydinskyi@gmail.com>",
    url="https://github.com/vbrydik/pyw2v2",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        "torch", 
        "torchaudio", 
        "librosa", 
        "transformers", 
        "datasets", 
        "easydict", 
        "PyYAML"
    ],
    keywords=['python'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)